import re
from clearskies.input_requirements.requirement import Requirement


class LettersDigits(Requirement):
    def configure(self):
        pass

    def check(self, model, data):
        if not data.get(self.column_name):
            return ""
        if not re.search("\d", data[self.column_name]):
            return f"{self.column_name} must contain numbers and letters, but does not contain any numbers."
        if not re.search("[a-zA-Z]", data[self.column_name]):
            return f"{self.column_name} must contain numbers and letters, but does not contain any letters."
        return ""



from clearskies.binding_config import BindingConfig
from .letters_digits import LettersDigits
from .letters_digits_special_characters import LettersDigitsSpecialCharacters
from .password_validation import PasswordValidation


def letters_digits():
    return BindingConfig(LettersDigits)


def letters_digits_special_characters(special_characters="!@#$%^&*()<>,.?~`"):
    return BindingConfig(LettersDigitsSpecialCharacters, special_characters=special_characters)


def password_validation():
    return BindingConfig(PasswordValidation)


__all__ = [
    "letters_digits",
    "letters_digits_special_characters",
    "LettersDigits",
    "LettersDigitsSpecialCharacters",
    "password_validation",
    "PasswordValidation",
]



from jwcrypto import jwk
import json

from .key_base import KeyBase


class CreateKey(KeyBase):
    _uuid = None

    def __init__(self, di, secrets, datetime, uuid):
        super().__init__(di, secrets, datetime)
        self._uuid = uuid

    def handle(self, input_output):
        # fetch the old keys
        private_keys = self.fetch_and_check_keys(self.configuration("path_to_private_keys"))
        public_keys = self.fetch_and_check_keys(self.configuration("path_to_public_keys"))
        self.check_for_inconsistencies(private_keys, public_keys)

        # make a new key
        key_id = str(self._uuid.uuid4())
        key = jwk.JWK.generate(
            kty=self.configuration("key_type"),
            size=self.configuration("key_size"),
            kid=key_id,
            alg=self.configuration("algorithm"),
            use="sig",
        )

        # and add it to our dictionaries
        private_keys[key_id] = {
            **json.loads(key.export_private()),
            "issue_date": self._datetime.datetime.now(self._datetime.timezone.utc).isoformat(),
        }
        public_keys[key_id] = {
            **json.loads(key.export_public()),
            "issue_date": self._datetime.datetime.now(self._datetime.timezone.utc).isoformat(),
        }

        self.save_keys(self.configuration("path_to_private_keys"), private_keys)
        self.save_keys(self.configuration("path_to_public_keys"), public_keys)

        return self.success(input_output, {"id": key_id})



import inspect
import json
from jwcrypto import jwk, jwt
from clearskies.handlers.exceptions import InputError
from clearskies.column_types import Audit
from .key_base import KeyBase
import datetime


class PasswordLogin(KeyBase):
    _configuration_defaults = {
        "user_model_class": "",
        "username_column_name": "email",
        "password_column_name": "password",
        "jwt_lifetime_seconds": 86400,
        "issuer": "",
        "audience": "",
        "path_to_private_keys": "",
        "path_to_public_keys": "",
        "key_cache_duration": 7200,
        "claims_callable": None,
        "claims_column_names": None,
        "input_error_callable": None,
        "login_check_callables": [],
        "audit": True,
        "audit_column_name": None,
        "audit_action_name_successful_login": "login",
        "audit_action_name_failed_login": "failed_login",
        "audit_action_name_account_locked": "account_lockout",
        "account_lockout": True,
        "account_lockout_failed_attempts_threshold": 10,
        "account_lockout_failed_attempts_period_minutes": 5,
        "users": None,
    }

    _required_configurations = [
        "issuer",
        "audience",
        "user_model_class",
        "path_to_private_keys",
        "path_to_public_keys",
    ]

    def __init__(self, di, secrets, datetime):
        super().__init__(di, secrets, datetime)
        self._columns = None

    def _check_configuration(self, configuration):
        super()._check_configuration(configuration)
        error_prefix = "Invalid configuation for handler " + self.__class__.__name__ + ":"
        for key in self._required_configurations:
            if not configuration.get(key):
                raise ValueError(f"{error_prefix} missing required configuration '{key}'")
        if "claims_callable" in configuration and not callable(configuration.get("claims_callable")):
            raise ValueError(f"{error_prefix} the provided 'claims_callable' configuration is not actually callable.")
        if "input_error_callable" in configuration and not callable(configuration.get("input_error_callable")):
            raise ValueError(
                f"{error_prefix} the provided 'input_error_callable' configuration is not actually callable."
            )

        user_model_class = configuration.get("user_model_class")
        if not inspect.isclass(user_model_class):
            raise ValueError(
                f"{error_prefix} 'user_model_class' should be a model class, but instead it is a '"
                + type(user_model_class)
                + "'"
            )
        if not getattr(user_model_class, "where"):
            raise ValueError(
                f"{error_prefix} 'user_model_class' should be a clearskies model class, but instead it is a '"
                + user_model_class.__name__
                + "'"
            )
        user_model = self._di.build(user_model_class)
        temporary_columns = user_model.columns()
        username_column_name = configuration.get("username_column_name", "email")
        password_column_name = configuration.get("password_column_name", "password")
        if username_column_name not in temporary_columns:
            raise ValueError(
                f"{error_prefix} the provided username column, '{username_column_name}', does not exist in the user model '{user_model_class.__name__}'"
            )
        if not temporary_columns[username_column_name].is_required:
            raise ValueError(
                f"{error_prefix} the provided username column, '{username_column_name}', in model '{user_model_class.__name__}' must be a required column."
            )
        if password_column_name not in temporary_columns:
            raise ValueError(
                f"{error_prefix} the provided password column, '{password_column_name}', does not exist in the user model '{user_model_class.__name__}'"
            )
        if not temporary_columns[password_column_name].is_required:
            raise ValueError(
                f"{error_prefix} the provided password column, '{password_column_name}', in model '{user_model_class.__name__}' must be a required column."
            )
        if not hasattr(temporary_columns[password_column_name], "validate_password"):
            raise ValueError(
                f"{error_prefix} the provided password column, '{password_column_name}', in model '{user_model_class.__name__}' does not implement the required 'validate_password' method.  You should double check to make sure it is using the 'clearskies_auth_server.columns.password column' type."
            )

        # we're getting columns twice, which is inefficient.  The reason why is because, for the columns object we actually use,
        # we need to set an override on the password column and set `for_login` to True.  I don't want to do this when I initially
        # fetch the columns above because, if the user passed in the wrong name for the password column, this would result in weird
        # and confusing errors.  Therefore, we get the columns above without any overrides, do the first round of user input validation, and then
        # we re-fetch the columns and set our overrides.
        self._columns = user_model.columns(overrides={password_column_name: {"for_login": True}})

        if configuration.get("claims_callable") and configuration.get("claims_column_names"):
            raise ValueError(
                f"{error_prefix} you set both 'claims_callable' and 'claims_column_names' but only one can be set."
            )
        if not configuration.get("claims_callable") and not configuration.get("claims_column_names"):
            raise ValueError(
                f"{error_prefix} you must set either 'claims_callable' or 'claims_column_names' but neither was set."
            )

        if configuration.get("claims_column_names"):
            claims_column_names = configuration["claims_column_names"]
            if not isinstance(claims_column_names, list):
                raise ValueError(
                    f"{error_prefix} 'claims_column_names' should be a list of column names, but instead has type "
                    + type(claims_column_names)
                )
            for column_name in claims_column_names:
                if column_name not in self._columns:
                    raise ValueError(
                        f"{error_prefix} a configured claim column, '{column_name}' does not exist in the user model"
                    )
                if not self._columns[column_name].is_readable:
                    raise ValueError(
                        f"{error_prefix} a configured claim column, '{column_name}' is not readable and so cannot be used in the claims"
                    )

        if configuration.get("account_lockout") and not configuration.get("audit"):
            raise ValueError(
                f"{error_prefix} 'account_lockout' is set to True but 'audit' is False.  You must enable auditing to turn on account lockouts."
            )

        if configuration.get("audit"):
            audit_column_name = configuration.get("audit_column_name")
            if audit_column_name not in self._columns:
                raise ValueError(
                    f"{error_prefix} 'audit_column_name' is '{audit_column_name}' but this column does not exist in the user model class, '{user_model_class.__name__}'"
                )
            if not isinstance(self._columns[audit_column_name], Audit):
                raise ValueError(
                    f"{error_prefix} 'audit_column_name' is '{audit_column_name}' but this column is not an audit column for the user model class, '{user_model_class.__name__}'"
                )

        if configuration.get("login_check_callables"):
            login_check_callables = configuration.get("login_check_callables")
            if not isinstance(login_check_callables, list):
                raise ValueError(
                    f"{error_prefix} 'login_check_callables' should be a list, but instead it has type '"
                    + type(login_check_callables)
                    + "'"
                )
            for index, login_check_callable in enumerate(login_check_callables):
                if not callable(login_check_callable):
                    raise ValueError(
                        f"{error_prefix} each entry in 'login_check_callables' should be a callable, but entry #{index} is not callable."
                    )

    def _get_audit_column(self, columns):
        audit_column = None
        for column in columns.values():
            if not isinstance(column, Audit):
                continue
            return column
        return None

    def apply_default_configuation(self, configuration):
        if not configuration.get("audit_column_name") and ("audit" not in configuration or configuration["audit"]):
            configuration["audit_column_name"] = self._get_audit_column(self._columns).name
        return super().apply_default_configuation(configuration)

    @property
    def users(self):
        return self._di.build(self.configuration("user_model_class"), cache=True)

    def handle(self, input_output):
        request_data = self.request_data(input_output)
        input_errors = self._find_input_errors(self.users, request_data, input_output)
        if input_errors:
            raise InputError(input_errors)

        username_column_name = self.configuration("username_column_name")
        password_column_name = self.configuration("password_column_name")
        password_column = self._columns[password_column_name]
        user = self.users.find(f"{username_column_name}=" + request_data[username_column_name])

        # no user found
        if not user.exists:
            return self.error(input_output, "Invalid username/password combination", 404)

        # account lockout
        if self.account_locked(user):
            self.audit(
                user,
                self.configuration("audit_action_name_account_locked"),
                data={
                    "reason": "Account Locked",
                },
            )
            minutes = self.configuration("account_lockout_failed_attempts_threshold")
            return self.error(
                input_output,
                f"Your account us under a {minutes} minute lockout due to too many failed login attempts",
                404,
            )

        # password not set
        if not user.get(password_column_name):
            self.audit(
                user,
                self.configuration("audit_action_name_failed_login"),
                data={
                    "reason": "Password not set - user is not configured for password login",
                },
            )
            return self.error(input_output, "Invalid username/password combination", 404)

        # invalid password
        if not password_column.validate_password(user, request_data[password_column_name]):
            self.audit(
                user,
                self.configuration("audit_action_name_failed_login"),
                data={
                    "reason": "Invalid password",
                },
            )
            return self.error(input_output, "Invalid username/password combination", 404)

        # developer-defined checks
        login_check_callables = self.configuration("login_check_callables")
        if login_check_callables:
            for login_check_callable in login_check_callables:
                response = self._di.call_function(
                    login_check_callable,
                    user=user,
                    request_data=request_data,
                    input_output=input_output,
                    **input_output.routing_data(),
                    **input_output.context_specifics(),
                )
                if response:
                    self.audit(
                        user,
                        self.configuration("audit_action_name_failed_login"),
                        data={
                            "reason": response,
                        },
                    )
                    return self.error(input_output, response, 404)

        self.audit(user, self.configuration("audit_action_name_successful_login"))
        signing_key = self.get_youngest_private_key(self.configuration("path_to_private_keys"), as_json=False)
        jwt_claims = self.get_jwt_claims(user)
        token = jwt.JWT(header={"alg": "RS256", "typ": "JWT"}, claims=jwt_claims)
        token.make_signed_token(signing_key)

        return input_output.respond(
            {
                "token": token.serialize(),
                "expires_at": jwt_claims["exp"],
            },
            200,
        )

    def account_locked(self, user):
        if not self.configuration("account_lockout"):
            return True

        threshold_time = datetime.datetime.now() - datetime.timedelta(
            minutes=self.configuration("account_lockout_failed_attempts_period_minutes")
        )
        audit_column_name = self.configuration("audit_column_name")
        failed_attempts = user.get(audit_column_name).where(
            "action=" + self.configuration("audit_action_name_failed_login")
        )
        failed_attempts = failed_attempts.where("created_at>" + threshold_time.strftime("%Y-%m-%d %H:%M:%S"))
        return len(failed_attempts) >= self.configuration("account_lockout_failed_attempts_threshold")

    def request_data(self, input_output, required=True):
        # make sure we don't drop any data along the way, because the input validation
        # needs to return an error for unexpected data.
        column_map = {
            self.configuration("username_column_name"): self.auto_case_column_name(
                self.configuration("username_column_name"), True
            ),
            self.configuration("password_column_name"): self.auto_case_column_name(
                self.configuration("password_column_name"), True
            ),
        }
        mapped_data = {}
        for key, value in input_output.request_data(required=required).items():
            mapped_data[column_map.get(key, key)] = value
        return mapped_data

    def _find_input_errors(self, model, request_data, input_output):
        input_errors = {}
        allowed_column_names = [
            self.configuration("username_column_name"),
            self.configuration("password_column_name"),
        ]
        for extra_column in set(request_data.keys()) - set(allowed_column_names):
            input_errors[extra_column] = "Input column '{extra_column}' is not an allowed column."
        for column_name in allowed_column_names:
            input_errors = {
                **input_errors,
                **self._columns[column_name].input_errors(model, request_data),
            }
        input_error_callable = self.configuration("input_error_callable")
        if input_error_callable:
            more_input_errors = self._di.call_function(
                input_error_callable,
                input_data=request_data,
                request_data=request_data,
                input_output=input_output,
                routing_data=input_output.routing_data(),
                authorization_data=input_output.get_authorization_data(),
            )
            if type(more_input_errors) != dict:
                raise ValueError(
                    "The input error callable, '"
                    + str(input_error_callable)
                    + "', did not return a dictionary as required"
                )
            input_errors = {
                **input_errors,
                **more_input_errors,
            }
        return input_errors

    def get_jwt_claims(self, user):
        if self.configuration("input_error_callable"):
            claims = self._di.call_function(user=user)
        else:
            claims = {
                claim_column: user.get(claim_column) for claim_column in self.configuration("claims_column_names")
            }

        now = self._datetime.datetime.now(self._datetime.timezone.utc)
        return {
            "aud": self.configuration("audience"),
            "iss": self.configuration("issuer"),
            "exp": int((now + datetime.timedelta(seconds=self.configuration("jwt_lifetime_seconds"))).timestamp()),
            **claims,
            "iat": int(now.timestamp()),
        }

    def audit(self, user, action_name, data=None):
        if not self.configuration("audit"):
            return
        self._columns[self.configuration("audit_column_name")].record(user, action_name, data=None)



from collections import OrderedDict
from clearskies.input_requirements import Requirement, required
from clearskies.column_types import string


class PasswordValidation(Requirement):
    def check(self, model, data):
        has_value = False
        has_some_value = False
        if self.column_name in data:
            has_some_value = True
            if type(data[self.column_name]) == str:
                has_value = bool(data[self.column_name].strip())
            else:
                has_value = bool(data[self.column_name])
        if has_value:
            return ""
        if model.exists and model[self.column_name] and not has_some_value:
            return ""
        return f"'{self.column_name}' is required."

    def additional_write_columns(self, is_create=False):
        # only needed on update
        if is_create:
            return {}

        return OrderedDict([string("validate_password", is_temporary=True)])



from jwcrypto import jwk
import json

from clearskies.handlers.exceptions import InputError
from .key_base import KeyBase


class DeleteOldestKey(KeyBase):
    def handle(self, input_output):
        # fetch the old keys
        private_keys = self.fetch_and_check_keys(self.configuration("path_to_private_keys"))
        public_keys = self.fetch_and_check_keys(self.configuration("path_to_public_keys"))
        self.check_for_inconsistencies(private_keys, public_keys)

        if len(private_keys) == 1:
            raise InputError("I'm cowardly refusing to delete the last key.  Sorry.")
        key_id = min(private_keys, key=lambda key_id: private_keys[key_id]["issue_date"])

        private_keys = {key: value for (key, value) in private_keys.items() if key != key_id}
        public_keys = {key: value for (key, value) in public_keys.items() if key != key_id}

        self.save_keys(self.configuration("path_to_private_keys"), private_keys)
        self.save_keys(self.configuration("path_to_public_keys"), public_keys)

        return self.success(input_output, {"id": key_id})



from clearskies.column_types import string, String
from clearskies.input_requirements import required
from passlib.context import CryptContext


class Password(String):
    _crypt_context = None

    my_configs = [
        "crypt_context",
        "crypt_context_string",
        "crypt_context_path",
        "require_repeat_password",
        "repeat_password_column_name",
        "for_login",
    ]

    crypt_config_names = [
        "crypt_context",
        "crypt_context_string",
        "crypt_context_path",
    ]

    def __init__(self, di):
        super().__init__(di)

    @property
    def is_readable(self):
        return False

    def _check_configuration(self, configuration):
        super()._check_configuration(configuration)
        count = 0
        for config_name in self.crypt_config_names:
            if configuration.get(config_name):
                count += 1
        if count > 1:
            raise ValueError(
                f"Error for column '{self.name}' in model '{self.model_class.__name__}': "
                + "you can only provide one of 'crypt_context', 'crypt_context_string', and 'crypt_context_path', "
                + "but more than one was found"
            )
        repeat_password_column_name = configuration.get("repeat_password_column_name", "repeat_password")
        if not isinstance(repeat_password_column_name, str):
            raise ValueError(
                f"Error for column '{self.name}' in model '{self.model_class.__name__}': "
                + "repeat_password_column_name should be a string, but instead I received a "
                + str(type(repeat_password_column_name))
            )

    def _finalize_configuration(self, configuration):
        configuration = super()._finalize_configuration(
            {"repeat_password_column_name": "repeat_password", **configuration}
        )
        found = False
        for config_name in self.crypt_config_names:
            if config_name in configuration:
                found = True
                break
        if not found:
            configuration["crypt_context"] = {
                "schemes": ["argon2"],
                "argon2__rounds": 5,
            }
        return configuration

    def configure(self, name, configuration, model_class):
        super().configure(name, configuration, model_class)
        if "crypt_context" in self.configuration:
            self._crypt_context = CryptContext(**self.config("crypt_context"))
        elif "crypt_context_string" in self.configuration:
            self._crypt_context = CryptContext.from_string(self.config("crypt_context_string"))
        else:
            self._crypt_context = CryptContext.from_path(self.config("crypt_context_path"))

    def check_input(self, model, data):
        if self.name not in data or not data[self.name]:
            return ""
        # if we're providing a password, then it must match the value in 'repeat_password'
        if self.config("for_login"):
            return ""
        repeat_password = data.get("repeat_password")
        password = data.get("password")
        return "Passwords did not match" if password != repeat_password else ""

    def pre_save(self, data, model):
        # if the password is being set to a non-value, then unset it
        if self.name in data and not data[self.name]:
            del data[self.name]
        elif data.get(self.name):
            data[self.name] = self._crypt_context.hash(data[self.name])
        return data

    def validate_password(self, user, password):
        hashed_password = user.get(self.name)
        if not hashed_password:
            return False

        if not self._crypt_context.verify(password, hashed_password):
            return False

        # yes, I understand that the crypt context has a `verify_and_update` flow for this, but
        # re-organizing the code to support that isn't worth the relatively minor efficiency gains.
        # The save process will automatically hash the password, so there isn't a flow to pass in
        # an already-hashed password.
        if self._crypt_context.needs_update(hashed_password):
            user.save({self.name: password})
        return True

    def additional_write_columns(self, is_create=False):
        if self.config("for_login"):
            return {}

        extra_columns = super().additional_write_columns(is_create=is_create)
        if self.config("require_repeat_password"):
            extra_columns.update([string("repeat_password", is_temporary=True)])
        return extra_columns



from clearskies.column_types import build_column_config
from .password import Password


def password(name, **kwargs):
    return build_column_config(name, Password, **kwargs)


__all__ = [
    "password",
    "Password",
]



from jwcrypto import jwk
import json

from .key_base import KeyBase


class Jwks(KeyBase):
    def handle(self, input_output):
        public_keys = self.fetch_and_check_keys(self.configuration("path_to_public_keys"))

        keys = [
            {
                "kid": key["kid"],
                "use": key["use"],
                "e": key["e"],
                "n": key["n"],
                "kty": key["kty"],
                "alg": key["alg"],
            }
            for key in public_keys.values()
        ]
        return self.success(input_output, keys)



import re
from .letters_digits import LettersDigits


class LettersDigitsSpecialCharacters(LettersDigits):
    special_characters = None

    def configure(self, special_characters="!@#$%^&*()<>,.?~`"):
        if type(special_characters) != str:
            raise ValueError(f"special_characters must be a string with the list of special characters to check for.")
        self.special_characters = special_characters

    def check(self, model, data):
        if not data.get(self.column_name):
            return ""
        error = super().check(model, data)
        if error:
            return error
        value = data[self.column_name]
        found = False
        for character in self.special_characters:
            if character not in value:
                continue
            found = True
            break

        if not found:
            return (
                f"{self.column_name} must contain at least one special character from the following list: "
                + ", ".join(list(self.special_characters))
            )
        return ""



from jwcrypto import jwk
import json
from clearskies.handlers.base import Base as HandlerBase


class KeyBase(HandlerBase):
    _secrets = None
    _datetime = None
    _key_cache = None
    _cache_time = None

    _configuration_defaults = {
        "path_to_public_keys": "",
        "path_to_private_keys": "",
        "algorithm": "RSA256",
        "key_type": "RSA",
        "key_size": 2048,
        "key_cache_duration": 7200,
    }

    def __init__(self, di, secrets, datetime):
        super().__init__(di)
        self._secrets = secrets
        self._datetime = datetime
        self._key_cache = {}

    def _check_configuration(self, configuration):
        super()._check_configuration(configuration)
        error_prefix = "Configuration error for %s:" % (self.__class__.__name__)
        for config_name in ["path_to_private_keys", "path_to_public_keys"]:
            if not configuration.get(config_name):
                raise ValueError(f"{error_prefix} the configuration value '{config_name}' is required but missing.")
        algorithm = configuration.get("algorithm")
        if algorithm and algorithm != "RSA256":
            raise ValueError("Currently only RSA256 is supported for the algorithm.")
        key_type = configuration.get("key_type")
        if key_type and key_type != "RSA":
            raise ValueError("Currently only RSA keys are supported.")

    def fetch_and_check_keys(self, path, use_cache=True):
        if use_cache and path in self._key_cache:
            cache_valid_time = self._datetime.datetime.now() - self._datetime.timedelta(
                seconds=self.configuration("cache_valid_time")
            )
            if self._key_cache[path]["cache_time"] > cache_valid_time:
                return self._key_cache[path]["key_data"]

        raw_data = self._secrets.get(path, silent_if_not_found=True)
        if not raw_data:
            return {}

        try:
            key_data = json.loads(raw_data)
        except json.JSONDecodeError as e:
            raise ValueError(
                f"I fetched the key data from '{path}'.  It should have been a JSON encoded object but it isn't JSON.  Sorry :("
            )

        actual_type = type(key_data)
        if actual_type != dict:
            raise ValueError(
                f"The key data stored in '{path}' should have been a dictionary but instead was a '{actual_type.__name__}'"
            )

        if use_cache:
            self._key_cache[path] = {"cache_time": self._datetime.datetime.now(), "key_data": key_data}

        # that's as far as we're going to get for now.
        return key_data

    def get_oldest_private_key(self, path, use_cache=True, as_json=True):
        keys = self.fetch_and_check_keys(path, use_cache=use_cache)
        oldest_key_id = min(keys, key=lambda key_id: keys[key_id]["issue_date"])
        return keys[oldest_key_id] if as_json else jwk.JWK(**keys[oldest_key_id])

    def get_youngest_private_key(self, path, use_cache=True, as_json=True):
        keys = self.fetch_and_check_keys(path, use_cache=use_cache)
        youngest_key_id = max(keys, key=lambda key_id: keys[key_id]["issue_date"])
        return keys[youngest_key_id] if as_json else jwk.JWK(**keys[youngest_key_id])

    def check_for_inconsistencies(self, private_keys, public_keys):
        """
        Checks that the public and private keys have the same set of keys.

        This exists because the private keys are stored in separate locations, so it's
        good to have a quick sanity check that they haven't grown out of sync.

        They are stored separately so that you can use separate permission controls
        for the private and public keys (if desired).

        The main thing we'll check for is to confirm that they have the same key ids
        """
        # the key ids are also the key in the dictionary
        private_key_ids = set(private_keys.keys())
        public_key_ids = set(public_keys.keys())
        extra_private_keys = private_key_ids - public_key_ids
        extra_public_keys = public_key_ids - private_key_ids
        if extra_private_keys:
            raise ValueError(
                "There are some private keys that don't have corresponding public keys.  Those are: '"
                + "', '".join(extra_private_keys)
                + "'.  You'll have to manually restore the missing key or delete the extra key."
            )
        if extra_public_keys:
            raise ValueError(
                "There are some public keys that don't have corresponding private keys.  Those are: '"
                + "', '".join(extra_public_keys)
                + "'.  You'll have to manually restore the missing key or delete the extra key."
            )

    def save_keys(self, path, keys):
        self._secrets.upsert(path, json.dumps(keys))



from jwcrypto import jwk
import json

from clearskies.handlers.exceptions import NotFound, InputError
from .key_base import KeyBase


class DeleteKey(KeyBase):
    def handle(self, input_output):
        # fetch the old keys
        private_keys = self.fetch_and_check_keys(self.configuration("path_to_private_keys"))
        public_keys = self.fetch_and_check_keys(self.configuration("path_to_public_keys"))
        self.check_for_inconsistencies(private_keys, public_keys)

        key_id = input_output.routing_data().get("key_id")
        if key_id not in private_keys:
            raise NotFound(f"Key '{key_id}' not found")
        if len(private_keys) == 1:
            raise InputError({"id": "I'm cowardly refusing to delete the last key.  Sorry."})

        private_keys = {key: value for (key, value) in private_keys.items() if key != key_id}
        public_keys = {key: value for (key, value) in public_keys.items() if key != key_id}

        self.save_keys(self.configuration("path_to_private_keys"), private_keys)
        self.save_keys(self.configuration("path_to_public_keys"), public_keys)

        return self.success(input_output, {"id": key_id})



import secrets
import inspect
from clearskies.handlers.exceptions import InputError
from clearskies.handlers.base import Base
from clearskies.column_types import Audit


class PasswordResetRequest(Base):
    _configuration_defaults = {
        "user_model_class": "",
        "username_column_name": "email",
        "email_column_name": "email",
        "reset_key_lifetime_seconds": 86400,
        "reset_key_column_name": "reset_key",
        "reset_expiration_column_name": "reset_key_expiration",
        "input_error_callable": None,
        "audit": True,
        "audit_column_name": None,
        "audit_action_name": "request_password_reset",
        "users": None,
    }

    _required_configurations = [
        "user_model_class",
    ]

    def __init__(self, di, datetime):
        super().__init__(di)
        self._columns = None
        self._datetime = datetime

    def _check_configuration(self, configuration):
        super()._check_configuration(configuration)
        error_prefix = "Invalid configuation for handler " + self.__class__.__name__ + ":"
        for key in self._required_configurations:
            if not configuration.get(key):
                raise ValueError(f"{error_prefix} missing required configuration '{key}'")

        user_model_class = configuration.get("user_model_class")
        if not inspect.isclass(user_model_class):
            raise ValueError(
                f"{error_prefix} 'user_model_class' should be a model class, but instead it is a '"
                + type(user_model_class)
                + "'"
            )
        if not getattr(user_model_class, "where"):
            raise ValueError(
                f"{error_prefix} 'user_model_class' should be a clearskies model class, but instead it is a '"
                + user_model_class.__name__
                + "'"
            )
        users = self._di.build(user_model_class)
        self._columns = users.columns()

        columns_to_check = [
            "username_column_name",
            "email_column_name",
            "reset_key_column_name",
            "reset_expiration_column_name",
        ]
        for config_name in columns_to_check:
            is_default = config_name not in configuration
            column_name = configuration.get(config_name, self._configuration_defaults.get(config_name))
            if column_name not in self._columns:
                if is_default:
                    raise ValueError(
                        f"{error_prefix} the configuration setting, '{config_name}' is not set and the deafult column name, '{column_name}', does not exist in the user model '{user_model_class.__name__}'"
                    )
                else:
                    raise ValueError(
                        f"{error_prefix} the provided column name for {config_name}, '{column_name}', does not exist in the user model '{user_model_class.__name__}'"
                    )
        lifetime = configuration.get("reset_key_lifetime_seconds")
        if lifetime and not isinstance(lifetime, int):
            raise ValueError(f"{error_prefix} the provided value for 'reset_key_lifetime_seconds' must be an integer.")

        if configuration.get("audit"):
            audit_column_name = configuration.get("audit_column_name")
            if audit_column_name not in self._columns:
                raise ValueError(
                    f"{error_prefix} 'audit_column_name' is '{audit_column_name}' but this column does not exist in the user model class, '{user_model_class.__name__}'"
                )
            if not isinstance(self._columns[audit_column_name], Audit):
                raise ValueError(
                    f"{error_prefix} 'audit_column_name' is '{audit_column_name}' but this column is not an audit column for the user model class, '{user_model_class.__name__}'"
                )

    def _get_audit_column(self, columns):
        audit_column = None
        for column in columns.values():
            if not isinstance(column, Audit):
                continue
            return column
        return None

    def apply_default_configuation(self, configuration):
        if not configuration.get("audit_column_name") and ("audit" not in configuration or configuration["audit"]):
            configuration["audit_column_name"] = self._get_audit_column(self._columns).name
        return super().apply_default_configuation(configuration)

    @property
    def users(self):
        return self._di.build(self.configuration("user_model_class"), cache=True)

    def handle(self, input_output):
        request_data = self.request_data(input_output)
        input_errors = self._find_input_errors(self.users, request_data, input_output)
        if input_errors:
            raise InputError(input_errors)

        username_column_name = self.configuration("username_column_name")
        user = self.users.find(f"{username_column_name}=" + request_data[username_column_name])

        # no user found.  Don't return data since that gives away if the user exists in the system.
        if not user.exists:
            return self.success(input_output, {})

        self.audit(user, self.configuration("audit_action_name"))
        user.save(
            {
                self.configuration("reset_key_column_name"): secrets.token_urlsafe(nbytes=32),
                self.configuration("reset_expiration_column_name"): self._datetime.datetime.now(
                    self._datetime.timezone.utc
                )
                + self._datetime.timedelta(seconds=self.configuration("reset_key_lifetime_seconds")),
            }
        )

        return self.success(input_output, {})

    def request_data(self, input_output, required=True):
        # make sure we don't drop any data along the way, because the input validation
        # needs to return an error for unexpected data.
        column_map = {
            self.configuration("username_column_name"): self.auto_case_column_name(
                self.configuration("username_column_name"), True
            ),
        }
        mapped_data = {}
        for key, value in input_output.request_data(required=required).items():
            mapped_data[column_map.get(key, key)] = value
        return mapped_data

    def _find_input_errors(self, model, request_data, input_output):
        input_errors = {}
        allowed_column_names = [
            self.configuration("username_column_name"),
        ]
        for extra_column in set(request_data.keys()) - set(allowed_column_names):
            input_errors[extra_column] = "Input column '{extra_column}' is not an allowed column."
        for column_name in allowed_column_names:
            input_errors = {
                **input_errors,
                **self._columns[column_name].input_errors(model, request_data),
            }
        input_error_callable = self.configuration("input_error_callable")
        if input_error_callable:
            more_input_errors = self._di.call_function(
                input_error_callable,
                input_data=request_data,
                request_data=request_data,
                input_output=input_output,
                routing_data=input_output.routing_data(),
                authorization_data=input_output.get_authorization_data(),
            )
            if type(more_input_errors) != dict:
                raise ValueError(
                    "The input error callable, '"
                    + str(input_error_callable)
                    + "', did not return a dictionary as required"
                )
            input_errors = {
                **input_errors,
                **more_input_errors,
            }
        return input_errors

    def audit(self, user, action_name, data=None):
        if not self.configuration("audit"):
            return
        self._columns[self.configuration("audit_column_name")].record(user, action_name, data=None)



from clearskies_auth_server import handlers
import clearskies


def key_manager(
    path_to_public_keys: str = "",
    path_to_private_keys: str = "",
    algorithm: str = None,
    key_type: str = None,
    key_size: int = None,
    authentication: clearskies.BindingConfig = None,
) -> clearskies.Application:
    if not path_to_public_keys:
        raise ValueError(
            "You must provide the path to the public keys in your secret manager when using the key_manager application"
        )
    if not path_to_private_keys:
        raise ValueError(
            "You must provide the path to the public keys in your secret manager when using the key_manager application"
        )
    handler_config = {
        "path_to_public_keys": path_to_public_keys,
        "path_to_private_keys": path_to_private_keys,
        "algorithm": algorithm,
        "key_type": key_type,
        "key_size": key_size,
    }
    handler_config = {key: value for (key, value) in handler_config.items() if value}

    routing_config = {
        "routes": [
            {
                "path": "",
                "handler_class": handlers.CreateKey,
                "handler_config": handler_config,
                "methods": ["POST"],
            },
            {
                "path": "",
                "handler_class": handlers.DeleteOldestKey,
                "handler_config": handler_config,
                "methods": ["DELETE"],
            },
            {
                "path": "",
                "handler_class": handlers.ListKeys,
                "handler_config": handler_config,
            },
            {
                "path": "{key_id}",
                "handler_class": handlers.DeleteKey,
                "handler_config": handler_config,
                "methods": ["DELETE"],
            },
        ],
    }
    if authentication:
        routing_config["authentication"] = authentication

    return clearskies.Application(
        clearskies.handlers.SimpleRouting,
        routing_config,
    )



from jwcrypto import jwk
import json

from .key_base import KeyBase


class ListKeys(KeyBase):
    def handle(self, input_output):
        # fetch the old keys
        private_keys = self.fetch_and_check_keys(self.configuration("path_to_private_keys"))
        public_keys = self.fetch_and_check_keys(self.configuration("path_to_public_keys"))
        self.check_for_inconsistencies(private_keys, public_keys)

        keys = [
            {
                "id": key["kid"],
                "algorithm": key["alg"],
                "issue_date": key["issue_date"],
            }
            for key in private_keys.values()
        ]
        return self.success(input_output, keys)



from .create_key import CreateKey
from .delete_key import DeleteKey
from .delete_oldest_key import DeleteOldestKey
from .list_keys import ListKeys
from .key_base import KeyBase
from .jwks import Jwks

# from .password_less_email_request_login import PasswordLessEmailRequestLogin
# from .password_less_validate_login import PasswordLessValidateLogin
from .password_login import PasswordLogin

__all__ = [
    "CreateKey",
    "DeleteKey",
    "DeleteOldestKey",
    "ListKeys",
    "KeyBase",
    "Jwks",
    "PasswordLogin",
]
